import os
import torch
import torch.nn.functional as F
from collections import OrderedDict
from torch.autograd import Variable
from itertools import chain as ichain
from .networks import define_Linear_Mapping, define_Linear_Reconstruction, define_Linear_Discriminator, define_Linear_Decomposer

__author__ = "Zhijian Yang"
__copyright__ = "Copyright 2019-2020 The CBICA & SBIA Lab"
__credits__ = ["Zhijian Yang"]
__license__ = "See LICENSE file"
__version__ = "0.1.0"
__maintainer__ = "Zhijian Yang"
__email__ = "zhijianyang@outlook.com"
__status__ = "Development"


#####sample from discrete uniform random variable and construct SUB variable. 
def sample_z_previous(real_X,index, npattern):
    Tensor = torch.FloatTensor
    z_idx = torch.empty(real_X.size(0), dtype=torch.long)
    z_idx[:] = index
    z = Tensor(real_X.size(0), npattern).fill_(0)
    zt_random = Variable(real_X.data.new(real_X.size(0),1).uniform_(0,1))
    z = z.scatter(1, z_idx.unsqueeze(1), zt_random)
    #zt_random = zt_random.scatter_(1, index.unsqueeze(1), 0)
    return z

def sample_z_later(pre,real_X, index,npattern):
    Tensor = torch.FloatTensor
    pre = pre[:,index:index+1]
    z_idx = torch.empty(real_X.size(0), dtype=torch.long)
    z_idx[:] = index
    z = Tensor(real_X.size(0), npattern).fill_(0)
    zt_random = Variable((1 - pre) * torch.rand(real_X.size(0),1) + pre)
    z = z.scatter(1, z_idx.unsqueeze(1), zt_random)
    return z

def sample_z_cn(real_X, npattern):
    z_random = Variable(real_X.data.new(real_X.size(0),npattern).uniform_(0,0.05))
    return z_random


def criterion_GAN(pred, target_is_real):
    if target_is_real:
        target_var = Variable(pred.data.new(pred.shape[0]).long().fill_(0.))
        loss=F.cross_entropy(pred, target_var)
    else:
        target_var = Variable(pred.data.new(pred.shape[0]).long().fill_(1.))
        loss = F.cross_entropy(pred, target_var)
    return loss

def criterion_orthogonal(change_batch_sum, npattern):
    change_batch_sum = torch.abs(change_batch_sum)
    change_batch_sum = change_batch_sum/(torch.norm(change_batch_sum,dim=1).view(npattern,1))
    matrix = torch.matmul(change_batch_sum, torch.transpose(change_batch_sum,0,1))
    return F.mse_loss(matrix,torch.eye(npattern))

def mono_loss(later, prev):
    return F.mse_loss(torch.max(torch.abs(prev)-torch.abs(later),torch.zeros_like(later)),torch.zeros_like(later))

class dotdict(dict):
    """dot.notation access to dictionary attributes"""
    __getattr__ = dict.get
    __setattr__ = dict.__setitem__
    __delattr__ = dict.__delitem__

class SurrealGAN(object):
    def __init__(self):
        self.opt = None

        ##### definition of all netwotks
        self.netMapping = None
        self.netReconstruction= None
        self.netDiscriminator = None

        ##### definition of all optimizers
        self.optimizer_M = None
        self.optimizer_D = None

        ##### definition of all criterions
        self.criterionGAN = criterion_GAN
        self.criterionChange = F.l1_loss
        self.criterionMono = mono_loss
        self.criterionRecons = F.mse_loss
        self.criterionOrtho = criterion_orthogonal


    def create(self, opt):

        self.opt = opt

        ## definition of all netwotks
        self.netMapping = define_Linear_Mapping(self.opt.nROI,self.opt.npattern)
        self.netReconstruction = define_Linear_Reconstruction(self.opt.nROI,self.opt.npattern)
        self.netDiscriminator = define_Linear_Discriminator(self.opt.nROI,self.opt.npattern)
        self.netDecomposer = define_Linear_Decomposer(self.opt.nROI,self.opt.npattern)

        ## definition of all optimizers
        self.optimizer_M = torch.optim.Adam(ichain(self.netMapping.parameters(),self.netDecomposer.parameters(),self.netReconstruction.parameters()),
                                        lr=self.opt.lr, betas=(self.opt.beta1, 0.999),weight_decay=2.5*1e-3)
        self.optimizer_D = torch.optim.Adam(ichain(self.netDiscriminator.parameters()),
                                            lr=self.opt.lr/5., betas=(self.opt.beta1, 0.999))


    def train_instance(self, x, real_y):
        z_pre = [sample_z_previous(x,i,self.opt.npattern) for i in range(self.opt.npattern)]
        z_pre_sum = torch.sum(torch.stack(z_pre),dim=0)
        z_later = [sample_z_later(z_pre[i],x,i,self.opt.npattern) for i in range(self.opt.npattern)]
        z_later_sum = torch.sum(torch.stack(z_later),dim=0)

        fake_y_pre_change = [self.netMapping.forward(x,z_pre[i]) for i in range(self.opt.npattern)]
        fake_y_later_change = [self.netMapping.forward(x,z_later[i]) for i in range(self.opt.npattern)]

        fake_y_pre_change_tensor = torch.stack(fake_y_pre_change)
        fake_y_later_change_tensor = torch.stack(fake_y_later_change)

        fake_y_pre = x + torch.sum(fake_y_pre_change_tensor, dim=0)
        fake_y_later = x + torch.sum(fake_y_later_change_tensor, dim=0)

        fake_y_change_batch_sum = torch.mean(torch.cat((fake_y_pre_change_tensor , fake_y_later_change_tensor),1),dim=1)
        

        z_cn = sample_z_cn(x, self.opt.npattern)
        fake_y_cn = self.netMapping.forward(x,z_cn) + x

        fake_y = fake_y_pre
        fake_change_decompose = fake_y_pre_change_tensor
        fake_z_decompose = torch.stack(z_pre)

        ## Discriminator loss
        pred_fake_y = self.netDiscriminator.forward(fake_y.detach())
        loss_D_fake_y = self.criterionGAN(pred_fake_y, False)
        pred_true_y = self.netDiscriminator.forward(real_y)
        loss_D_true_y = self.criterionGAN(pred_true_y, True)
        loss_D= 0.5* (loss_D_fake_y + loss_D_true_y)

        ## update weights of discriminator
        self.optimizer_D.zero_grad()
        loss_D.backward()
        gnorm_D = torch.nn.utils.clip_grad_norm_(self.netDiscriminator.parameters(), self.opt.max_gnorm)
        self.optimizer_D.step()

        ## Mapping related loss
        pred_fake_y = self.netDiscriminator.forward(fake_y)
        loss_mapping = self.criterionGAN(pred_fake_y, True)

        decomposed_fake_y = self.netDecomposer.forward(fake_y)

        reconst_zt = [self.netReconstruction.forward(decomposed_fake_y[i]) for i in range(self.opt.npattern)]

        recons_loss = self.criterionRecons(torch.stack(reconst_zt), fake_z_decompose)

        decompose_loss = self.criterionRecons(torch.stack(decomposed_fake_y), fake_change_decompose)

        change_loss= self.criterionChange(fake_y, x)

        mono_loss = self.criterionMono(fake_y_later_change_tensor, fake_y_pre_change_tensor)

        orthogonal_loss = self.criterionOrtho(fake_y_change_batch_sum,self.opt.npattern)

        cn_loss = self.criterionChange(fake_y_cn, x)

        loss_G = loss_mapping+self.opt.lam*orthogonal_loss+ self.opt.zeta*recons_loss +self.opt.kappa*decompose_loss + self.opt.gamma*change_loss + self.opt.mu*mono_loss + self.opt.eta*cn_loss #7


        ## update weights of mapping/reconstruction/decomposer function
        self.optimizer_M.zero_grad()
        loss_G.backward()
        gnorm_M = torch.nn.utils.clip_grad_norm_(self.netMapping.parameters(), self.opt.max_gnorm)
        self.optimizer_M.step()

        ## perform weight clipping

        for p in self.netMapping.parameters():
            p.data.clamp_(-self.opt.lipschitz_k, self.opt.lipschitz_k)
        for p in self.netReconstruction.parameters():
            p.data.clamp_(-self.opt.lipschitz_k, self.opt.lipschitz_k)
        for p in self.netDecomposer.parameters():
            p.data.clamp_(-self.opt.lipschitz_k, self.opt.lipschitz_k)


        ## Return dicts
        losses=OrderedDict([('Discriminator_loss', loss_D.item()),('Mapping_loss', loss_mapping.item()),('loss_change', change_loss.item()),('loss_orthogonal', orthogonal_loss.item())
            ,('loss_recons', recons_loss.item()),('loss_mono', 100*mono_loss.item()),('loss_cn', cn_loss.item()),('loss_decompose', decompose_loss.item())])

        
        return losses


    
    ## return decomposed changes of each dimension given PT data
    def decompose(self,real_y):
        decompose_y=self.netDecomposer.forward(real_y)
        return [decompose_y[i].detach().numpy() for i in range(self.opt.npattern)]

    ## return rindices given PT data
    def predict_rindices(self,real_y):
        decompose_y=self.netDecomposer.forward(real_y)
        reconst_zt_decompose = [self.netReconstruction.forward(decompose_y[i]) for i in range(self.opt.npattern)]
        t = torch.sum(torch.stack(reconst_zt_decompose),dim=0)
        return t.detach().numpy()

    ## return generated patient data with given zub variable
    def predict_Y(self, x, z):
        return self.netMapping.forward(x, z)

    ## save checkpoint    
    def save(self, save_dir, chk_name):
        chk_path = os.path.join(save_dir, chk_name)
        checkpoint = {
            'netMapping':self.netMapping.state_dict(),
            'netDiscriminator':self.netDiscriminator.state_dict(),
            'optimizer_D':self.optimizer_D.state_dict(),
            'optimizer_M':self.optimizer_M.state_dict(),
            'netReconstruction':self.netReconstruction.state_dict(),
            'netDecomposer':self.netDecomposer.state_dict(),
        }
        checkpoint.update(self.opt)
        torch.save(checkpoint, chk_path)

    def load_opt(self,checkpoint):
        self.opt = dotdict({})
        for key in checkpoint.keys():
            if key not in ['netMapping','netDiscriminator','netReconstruction','netDecomposer','optimizer_M','optimizer_D']:
                self.opt[key] = checkpoint[key]
        

    ## load trained model
    def load(self, chk_path):
        checkpoint = torch.load(chk_path)
        self.load_opt(checkpoint)
        ##### definition of all netwotks
        self.netMapping = define_Linear_Mapping(self.opt.nROI,self.opt.npattern)
        self.netReconstruction = define_Linear_Reconstruction(self.opt.nROI,self.opt.npattern)
        self.netDiscriminator = define_Linear_Discriminator(self.opt.nROI,self.opt.npattern)
        self.netDecomposer = define_Linear_Decomposer(self.opt.nROI,self.opt.npattern)

        ##### definition of all optimizers
        self.optimizer_M = torch.optim.Adam(ichain(self.netMapping.parameters(),self.netDecomposer.parameters(),self.netReconstruction.parameters()),
                                        lr=self.opt.lr, betas=(self.opt.beta1, 0.999),weight_decay=2.5*1e-3)
        self.optimizer_D = torch.optim.Adam(ichain(self.netDiscriminator.parameters()),
                                        lr=self.opt.lr/5., betas=(self.opt.beta1, 0.999))
        
        self.netMapping.load_state_dict(checkpoint['netMapping'])
        self.netDiscriminator.load_state_dict(checkpoint['netDiscriminator'])
        self.netReconstruction.load_state_dict(checkpoint['netReconstruction'])
        self.netDecomposer.load_state_dict(checkpoint['netDecomposer'])
        self.optimizer_D.load_state_dict(checkpoint['optimizer_D'])
        self.optimizer_M.load_state_dict(checkpoint['optimizer_M'])
        self.load_opt(checkpoint)
            


        



import numpy as np
import scipy
import pandas as pd
import itertools
import torch 
from sklearn.preprocessing import minmax_scale
from sklearn.linear_model import LinearRegression
from sklearn.cluster import KMeans
from sklearn import metrics
from torch.autograd import Variable
from .data_loading import PTIterator, CNIterator, val_PT_construction, val_CN_construction
from lifelines.utils import concordance_index

__author__ = "Zhijian Yang"
__copyright__ = "Copyright 2019-2020 The CBICA & SBIA Lab"
__credits__ = ["Zhijian Yang"]
__license__ = "See LICENSE file"
__version__ = "0.1.0"
__maintainer__ = "Zhijian Yang"
__email__ = "zhijianyang@outlook.com"
__status__ = "Development"

def Covariate_correction(cn_data,cn_cov,pt_data,pt_cov):
    """
    Eliminate the confound of covariate, such as age and sex, from the disease-based changes.
    :param cn_data: array, control data
    :param cn_cov: array, control covariates
    :param pt_data: array, patient data
    :param pt_cov: array, patient covariates
    :return: corrected control data & corrected patient data
    """
    min_cov = np.amin(cn_cov, axis=0)
    max_cov = np.amax(cn_cov, axis=0)
    pt_cov = (pt_cov-min_cov)/(max_cov-min_cov)
    cn_cov = (cn_cov-min_cov)/(max_cov-min_cov)
    beta = np.transpose(LinearRegression().fit(cn_cov, cn_data).coef_)
    corrected_cn_data = cn_data-np.dot(cn_cov,beta)
    corrected_pt_data = pt_data-np.dot(pt_cov,beta)
    correction_variables = {'max_cov':max_cov,'min_cov':min_cov,'beta':beta}
    return corrected_cn_data, corrected_pt_data, correction_variables

def Data_normalization(cn_data,pt_data):
    """
    Normalize all data with respect to control data to ensure a mean of 1 and std of 0.1 
    among CN participants for each ROI
    :param cn_data: array, control data
    :param pt_data: array, patient data
    :return: normalized control data & normalized patient data
    """
    cn_mean = np.mean(cn_data,axis=0)
    cn_std = np.std(cn_data,axis=0)
    normalized_cn_data = 1+(cn_data-cn_mean)/(10*cn_std)
    normalized_pt_data = 1+(pt_data-cn_mean)/(10*cn_std)
    normalization_variables = {'cn_mean':cn_mean, 'cn_std':cn_std}
    return normalized_cn_data, normalized_pt_data, normalization_variables

def apply_covariate_correction(data, covariate, correction_variables):
    covariate = (covariate-correction_variables['min_cov'])/(correction_variables['max_cov']-correction_variables['min_cov'])
    corrected_data = data-np.dot(covariate,correction_variables['beta'])
    return corrected_data

def apply_data_normalization(data, normalization_variables):
    normalized_data = 1+(data-normalization_variables['cn_mean'])/(10*normalization_variables['cn_std'])
    return normalized_data

def parse_train_data(data, covariate, random_seed, data_fraction, batch_size):
    cn_data = data.loc[data['diagnosis'] == -1].drop(['participant_id','diagnosis'], axis=1).values
    pt_data = data.loc[data['diagnosis'] == 1].drop(['participant_id','diagnosis'], axis=1).values
    correction_variables = None
    if covariate is not None:
        cn_cov = covariate.loc[covariate['diagnosis'] == -1].drop(['participant_id', 'diagnosis'], axis=1).values
        pt_cov = covariate.loc[covariate['diagnosis'] == 1].drop(['participant_id','diagnosis'], axis=1).values
        cn_data,pt_data, correction_variables = Covariate_correction(cn_data,cn_cov,pt_data,pt_cov)
    normalized_cn_data, normalized_pt_data, normalization_variables = Data_normalization(cn_data,pt_data)
    cn_train_dataset = CNIterator(normalized_cn_data, random_seed, data_fraction, batch_size)
    pt_train_dataset = PTIterator(normalized_pt_data, random_seed, data_fraction, batch_size)
    return cn_train_dataset, pt_train_dataset, correction_variables, normalization_variables

def parse_validation_data(data, covariate, correction_variables, normalization_variables):
    cn_data = data.loc[data['diagnosis'] == -1].drop(['participant_id','diagnosis'], axis=1).values
    pt_data = data.loc[data['diagnosis'] == 1].drop(['participant_id','diagnosis'], axis=1).values
    if cn_data.shape[0] != 0:
        if correction_variables is not None:
            cn_cov = covariate.loc[covariate['diagnosis'] == -1].drop(['participant_id', 'diagnosis'], axis=1).values
            pt_cov = covariate.loc[covariate['diagnosis'] == 1].drop(['participant_id','diagnosis'], axis=1).values
            cn_data = apply_covariate_correction(cn_data, cn_cov, correction_variables)
            pt_data = apply_covariate_correction(pt_data, pt_cov, correction_variables)
        normalized_cn_data = apply_data_normalization(cn_data, normalization_variables)
        normalized_pt_data = apply_data_normalization(pt_data, normalization_variables)
        cn_eval_dataset = val_CN_construction(normalized_cn_data).load()
        pt_eval_dataset = val_PT_construction(normalized_pt_data).load()
        return cn_eval_dataset, pt_eval_dataset
    else:
        if correction_variables is not None:
            pt_cov = covariate.loc[covariate['diagnosis'] == 1].drop(['participant_id','diagnosis'], axis=1).values
            pt_data = apply_covariate_correction(pt_data, pt_cov, correction_variables)
        normalized_pt_data = apply_data_normalization(pt_data, normalization_variables)
        pt_eval_dataset = val_PT_construction(normalized_pt_data).load()
        return pt_eval_dataset






import sys
import os
import numpy as np
import itertools
from sklearn import metrics
from .model import SurrealGAN
from .utils import parse_validation_data
from .training import Surreal_GAN_train
from lifelines.utils import concordance_index

__author__ = "Zhijian Yang"
__copyright__ = "Copyright 2019-2020 The CBICA & SBIA Lab"
__credits__ = ["Zhijian Yang"]
__license__ = "See LICENSE file"
__version__ = "0.0.1"
__maintainer__ = "Zhijian Yang"
__email__ = "zhijianyang@outlook.com"
__status__ = "Development"

def calculate_pair_wise_c_index(r1,r2,npattern):
	# function for calculating pair-wise c-index between two saved models
	order_permutation = list(itertools.permutations(range(npattern)))
	c_indices = [ 0 for _ in range(npattern)]
	best_order = range(npattern)
	for i in range(len(order_permutation)):
		order_c_indices = [ 0 for _ in range(npattern)]
		for j in range(npattern):
			order_c_indices[j] = concordance_index(r1[:,j],r2[:,order_permutation[i][j]])
		if np.mean(order_c_indices) > np.mean(c_indices):
			c_indices = order_c_indices
			best_order = order_permutation[i]
	pairs = list(itertools.combinations(range(npattern),2))
	pair_wise_c_indices = 0
	for i in range(len(pairs)):
		pair_wise_c_indices += concordance_index(r1[:,pairs[i][0]]-r1[:,pairs[i][1]],r2[:,best_order[pairs[i][0]]]-r2[:,best_order[pairs[i][1]]])
	return pair_wise_c_indices/len(pairs), np.mean(c_indices)


def calculate_group_compare_c_index(prediction_rindices,npattern):
	# function for calculating pattern c indices among a groups of predicted r-indices
	all_diff_c_indices = []
	all_pattern_c_indices = []
	for i in range(len(prediction_rindices)):
		local_pattern_index = []
		local_diff_index = []
		for j in range(len(prediction_rindices)):
			if i!=j:
				diff_c_indices, patt_c_indices = calculate_pair_wise_c_index(prediction_rindices[i],prediction_rindices[j],npattern)
				local_diff_index.append(diff_c_indices)
				local_pattern_index.append(patt_c_indices)
		all_diff_c_indices.append(np.mean(local_diff_index))
		all_pattern_c_indices.append(np.mean(local_pattern_index))
	return all_diff_c_indices, all_pattern_c_indices

def apply_saved_model(model_dir, data, covariate=None):
	"""
	Function used for derive representation results from one saved model
	Args:
		model_dir: string, path to the saved data
		data, data_frame, dataframe with same format as training data. PT data can be any samples in or out of the training set.
		covariate, data_frame, dataframe with same format as training covariate. PT data can be any samples in or out of the training set.
	Returns: R-indices

	"""
	data = data[data['diagnosis']==1]
	if covariate is not None:
		covariate = covariate[covariate['diagnosis']==1]
	model = SurrealGAN()
	model.load(model_dir)
	validation_data = parse_validation_data(data, covariate,model.opt.correction_variables,model.opt.normalization_variables)
	model.predict_rindices(validation_data)
	return model.predict_rindices(validation_data)

def representation_result(model_dirs, npattern, data, final_saving_epoch, saving_freq, covariate=None):
	"""
	Function used for derive representation results from several saved models
	Args:
		model_dirs: list, list of dirs of all saved models
		npattern: int, number of pre-defined patterns
		data, data_frame, dataframe with same format as training data. 
		covariate, data_frame, dataframe with same format as training covariate. 
		final_saving_epoch: int, epoch number from which the last model will be saved and model training will be stopped if saving criteria satisfied
	Returns: R-indices, Pattern c-indices between the selected repetition and all other repetitionss, Pattern c-indices among all repetitions, path to the final selected model used for deriving R-indices

	"""
	best_iteration_prediction_rindices = []
	best_pattern_c_index_mean = 0
	best_pattern_agr_index = []
	best_pattern_diff_agr_index = []
	best_epoch = 0

	if final_saving_epoch % saving_freq == 0:
		save_epoch = [i * saving_freq for i in range(2,final_saving_epoch//saving_freq+1)]
	else:
		save_epoch = [i * saving_freq for i in range(2,final_saving_epoch//saving_freq+1)]+[final_saving_epoch]
	if len(model_dirs) > 9:
		for epoch in save_epoch:
			all_prediction_rindices = []
			for models in model_dirs:
				model = SurrealGAN()
				model.load(os.path.join(models, str(epoch)+'_epoch_model'))
				validation_data = parse_validation_data(data, covariate,model.opt.correction_variables,model.opt.normalization_variables)[1]
				all_prediction_rindices.append(model.predict_rindices(validation_data))
			pattern_diff_agr_index, pattern_agr_index = calculate_group_compare_c_index(all_prediction_rindices,npattern)
			if np.mean(pattern_agr_index)+np.mean(pattern_diff_agr_index) > best_pattern_c_index_mean:
				best_pattern_c_index_mean = np.mean(pattern_agr_index)+np.mean(pattern_diff_agr_index)
				best_pattern_agr_index = pattern_agr_index
				best_pattern_diff_agr_index = pattern_diff_agr_index
				best_iteration_prediction_rindices = all_prediction_rindices
				best_model_dir = os.path.join(model_dirs[pattern_agr_index.index(max(pattern_agr_index))], str(epoch)+'_epoch_model')
	
	else:
		raise Exception("At least 10 trained models are required (repetition number need to be at least 5)")
	max_index = best_pattern_diff_agr_index.index(max(best_pattern_diff_agr_index))
	return np.array(best_iteration_prediction_rindices[max_index]), best_pattern_agr_index[max_index], best_pattern_diff_agr_index[max_index], best_pattern_agr_index, best_pattern_diff_agr_index, best_model_dir


def repetitive_representation_learning(data, npattern, repetition, fraction, final_saving_epoch, max_epoch, output_dir, mono_loss_threshold=0.006, saving_freq = 1000,\
		recons_loss_threshold=0.003, covariate=None, lam=0.2, zeta=80, kappa=80, gamma=6, mu=500, eta=6, batchsize=100, lipschitz_k = 0.5, verbose = False, \
		beta1 = 0.5, lr = 0.0008, max_gnorm = 100, eval_freq = 50, save_epoch_freq = 5, start_repetition = 0, stop_repetition = None):
	"""
	Args:
		data: dataframe, dataframe file with all ROI (input features) The dataframe contains
		the following headers: "
								 "i) the first column is the participant_id;"
								 "iii) the second column should be the diagnosis;"
								 "The following column should be the extracted features. e.g., the ROI features"
		covariate: dataframe, not required; dataframe file with all confounding covariates to be corrected. The dataframe contains
		the following headers: "
								 "i) the first column is the participant_id;"
								 "iii) the second column should be the diagnosis;"
								 "The following column should be all confounding covariates. e.g., age, sex"
		npattern: int, number of defined patterns
		repetition: int, number of repetition of training process
		fraction: float, fraction of data used for training in each repetition
		final_saving_epoch: int, epoch number from which the last model will be saved and model training will be stopped if saving criteria satisfied
		max_epoch: int, maximum trainig epoch: training will stop even if criteria not satisfied.
		output_dir: str, the directory underwhich model and results will be saved
		mono_loss_threshold: float, chosen mono_loss theshold for stopping criteria
		recons_loss_threshold: float, chosen recons_loss theshold for stopping criteria
		lam: int, hyperparameter for orthogonal_loss
		zeta: int, hyperparameter for recons_loss
		kappa: int, hyperparameter for decompose_loss
		gamma: int, hyperparameter for change_loss
		mu: int, hyperparameter for mono_loss
		eta: int, hyperparameter for cn_loss
		batchsize: int, batck size for training procedure
		lipschitz_k: float, hyper parameter for weight clipping of transformation and reconstruction function
		verbose: bool, choose whether to print out training procedure
		beta1: float, parameter of ADAM optimization method
		lr: float, learning rate
		max_gnorm: float, maximum gradient norm for gradient clipping
		eval_freq: int, the frequency at which the model is evaluated during training procedure
		save_epoch_freq: int, the frequency at which the model is saved during training procedure
		start_repetition; int, indicate the last saved repetition index,
							  used for restart previous half-finished repetition training or for parallel training; set defaultly to be 0 indicating a new repetition training process
		stop_repetition: int, indicate the index of repetition at which the training process early stop,
							  used for stopping repetition training process eartly and resuming later or for parallel training; set defaultly to be None and repetition training will not stop till the end
		
	Returns: clustering outputs.

	"""
	print('Start Surreal-GAN for semi-supervised representation learning')

	Surreal_GAN_model = Surreal_GAN_train(npattern, final_saving_epoch, max_epoch, recons_loss_threshold, mono_loss_threshold, \
		lam=lam, zeta=zeta, kappa=kappa, gamma=gamma, mu=mu, eta=eta, batchsize=batchsize, \
		lipschitz_k = lipschitz_k, beta1 = beta1, lr = lr, max_gnorm = max_gnorm, eval_freq = eval_freq, save_epoch_freq = save_epoch_freq, saving_freq = saving_freq)

	if stop_repetition == None:
		stop_repetition = repetition
	for i in range(start_repetition, stop_repetition):
		print('****** Starting training of Repetition '+str(i)+" ******")
		converge = Surreal_GAN_model.train(data, covariate, output_dir, random_seed=i, data_fraction = fraction, verbose = verbose)
		while not converge:
			print("****** Model not converged at max interation, Start retraining ******")
			converge = Surreal_GAN_model.train(data, covariate, output_dir, random_seed=i, data_fraction = fraction, verbose = verbose)

	saved_models = [os.path.join(os.path.join(output_dir,'repetition'+str(i)))  for i in range(repetition)]
	
	r_indices, selected_model_pattern_agr_index, selected_model_pattern_diff_agr_index, pattern_agr_cindex, pattern_diff_agr_cindex, selected_model_dir = representation_result(saved_models, npattern, data, final_saving_epoch, saving_freq, covariate = covariate)
	
	pt_data = data.loc[data['diagnosis'] == 1][['participant_id','diagnosis']]

	for i in range(npattern):
		pt_data['r'+str(i+1)] = r_indices[:,i]

	pt_data["path to selected model"] = [selected_model_dir]+['' for _ in range(r_indices.shape[0]-1)]
	pt_data["selected model pattern-agr-index"] = ["%.3f" %(selected_model_pattern_agr_index)]+['' for _ in range(r_indices.shape[0]-1)]
	pt_data["pattern-agr-index" ] = ["%.3f +- %.3f" %(np.mean(pattern_agr_cindex), np.std(pattern_agr_cindex))]+['' for _ in range(r_indices.shape[0]-1)]
	pt_data["selected model pattern-diff-agr-index"] = ["%.3f" %(selected_model_pattern_diff_agr_index)]+['' for _ in range(r_indices.shape[0]-1)]
	pt_data["pattern-diff-agr-index" ] = ["%.3f +- %.3f" %(np.mean(pattern_diff_agr_cindex), np.std(pattern_diff_agr_cindex))]+['' for _ in range(r_indices.shape[0]-1)]
	
	pt_data.to_csv(os.path.join(output_dir,'representation_result.csv'), index = False)
	print('****** Surreal-GAN Representation Learning finished ******')



import torch
import torch.nn as nn

__author__ = "Zhijian Yang"
__copyright__ = "Copyright 2019-2020 The CBICA & SBIA Lab"
__credits__ = ["Zhijian Yang"]
__license__ = "See LICENSE file"
__version__ = "0.1.0"
__maintainer__ = "Zhijian Yang"
__email__ = "zhijianyang@outlook.com"
__status__ = "Development"

class TwoInputModule(nn.Module):
    def forward(self, input1, input2):
        raise NotImplementedError


class TwoInputSequential(nn.Sequential, TwoInputModule):
    def __init__(self, *args):
        super(TwoInputSequential, self).__init__(*args)

    def forward(self, input1, input2):
        ## overloads forward function in parent class
        for module in self._modules.values():
            if isinstance(module, TwoInputModule): #check whether it is twoinputmodule
                input1 = module.forward(input1, input2)
            else:
                input1 = module.forward(input1)
        return input1


class Sub_Adder(TwoInputModule):
    def __init__(self, x_dim, z_dim):
        super(Sub_Adder, self).__init__()
        self.add_noise = nn.Sequential(
            nn.Linear(z_dim, x_dim),
        )

    def forward(self, input, noise):
        #multiplier = self.add_noise.forward(noise)
        multiplier = torch.sigmoid(self.add_noise.forward(noise))
        return input*multiplier








import torch
import numpy as np

__author__ = "Zhijian Yang"
__copyright__ = "Copyright 2019-2020 The CBICA & SBIA Lab"
__credits__ = ["Zhijian Yang"]
__license__ = "See LICENSE file"
__version__ = "0.1.0"
__maintainer__ = "Zhijian Yang"
__email__ = "zhijianyang@outlook.com"
__status__ = "Development"

class PTIterator(object):
    def __init__(self, PT_data, random_seed, fraction, batch_size):
        super(PTIterator, self).__init__()
        np.random.seed(random_seed)
        indices = np.random.choice(PT_data.shape[0], int(fraction*PT_data.shape[0]), replace=False)
        self.data = PT_data.astype('float32')[indices]
        self.num_samples = self.data.shape[0]
        self.batch_size = batch_size
        self.n_batches = int(self.num_samples / self.batch_size)
        self.reset()

    def __iter__(self):
        return self

    def reset(self):
        self.data_indices = np.random.permutation(self.num_samples)
        self.batch_idx = 0

    def __next__(self):
        if self.batch_idx == self.n_batches-1:
            self.reset()
            raise StopIteration

        idx = self.batch_idx * self.batch_size
        chosen_indices = self.data_indices[idx:idx+self.batch_size]

        self.batch_idx += 1

        return {'y': torch.from_numpy(self.data[chosen_indices])}

    def __len__(self):
        return self.num_samples

class CNIterator(object):
    def __init__(self, CN_data, random_seed, fraction, batch_size):
        super(CNIterator, self).__init__()
        np.random.seed(random_seed)
        indices = np.random.choice(CN_data.shape[0], int(fraction*CN_data.shape[0]), replace=False)
        self.data = CN_data.astype('float32')[indices]
        self.num_samples = self.data.shape[0]
        self.batch_size = batch_size
        self.n_batches = int(self.num_samples / self.batch_size)
        self.reset()

    def __iter__(self):
        return self

    def reset(self):
        self.data_indices = np.random.permutation(self.num_samples)
        self.batch_idx = 0

    def next(self):
        if self.batch_idx == self.n_batches:
            self.reset()
            #raise StopIteration

        idx = self.batch_idx * self.batch_size
        chosen_indices = self.data_indices[idx:idx+self.batch_size]

        self.batch_idx += 1

        return {'x': torch.from_numpy(self.data[chosen_indices])}

    def __len__(self):
        return self.num_samples

class val_PT_construction(object):
    def __init__(self, PT_data, **kwargs):
        super(val_PT_construction, self).__init__()
        self.data = PT_data.astype('float32')

    def load(self):
        return torch.from_numpy(self.data)

    def __len__(self):
        return self.num_samples

class val_CN_construction(object):
    def __init__(self, CN_data, **kwargs):
        super(val_CN_construction, self).__init__()
        self.data = CN_data.astype('float32')

    def load(self):
        return torch.from_numpy(self.data)

    def __len__(self):
        return self.num_samples




import torch.nn as nn
import torch.nn.functional as F
from .modules import TwoInputSequential, Sub_Adder

__author__ = "Zhijian Yang"
__copyright__ = "Copyright 2019-2020 The CBICA & SBIA Lab"
__credits__ = ["Zhijian Yang"]
__license__ = "See LICENSE file"
__version__ = "0.1.0"
__maintainer__ = "Zhijian Yang"
__email__ = "zhijianyang@outlook.com"
__status__ = "Development"

###############################################################################
# Functions
###############################################################################

def weights_init(m):
    classname = m.__class__.__name__
    if classname.find('Linear') != -1:
        m.weight.data.normal_(0, 0.1)

def define_Linear_Mapping(nROI,nPattern):
    netG = LMappingGenerator(nPattern, nROI)
    netG.apply(weights_init)
    return netG

def define_Linear_Discriminator(nROI,nPattern):
    netD=LDiscriminator(nROI,nPattern)
    netD.apply(weights_init)
    return netD

def define_Linear_Reconstruction(nROI,nPattern):
    netC = LEncoder(nPattern,nROI)
    netC.apply(weights_init)
    return netC

def define_Linear_Decomposer(nROI,nPattern):
    netD = LDecompose(nPattern, nROI)
    netD.apply(weights_init)
    return netD

##############################################################################
# Network Classes
##############################################################################

class LMappingGenerator(nn.Module):
    def __init__(self, nPattern, nROI, product_layer=Sub_Adder):
        super(LMappingGenerator, self).__init__()
        model=[]
        def block(in_layer, out_layer, normalize=False):
            layers = [nn.Linear(in_layer, out_layer,bias=False)]
            if normalize:
                layers.append(nn.BatchNorm1d(out_layer, 0.8))
            layers.append(nn.LeakyReLU(0.2, inplace=True))
            return layers
        model+=block(nROI,int(nROI/2))+block(int(nROI/2), int(nROI/4))
        model.append(product_layer(int(nROI/4),nPattern))
        model+=block(int(nROI/4),int(nROI/2))+block(int(nROI/2),nROI)
        #model.append(nn.Linear(nROI, nROI,bias=False))
        self.model = TwoInputSequential(*model)
    def forward(self, input_x,input_z):
        return self.model(input_x,input_z)

class LEncoder(nn.Module):
    def __init__(self, nPattern, nROI):
        super(LEncoder, self).__init__()
        model = []
        self.nPattern = nPattern
        def block(in_layer, out_layer, normalize=True):
            layers=[]
            #layers.append(nn.BatchNorm1d(in_layer, 0.8))
            layers.append(nn.LeakyReLU(0.2, inplace=True))
            layers.append(nn.Linear(in_layer, out_layer))
            return layers
        model.append(nn.Linear(nROI, int(nROI/2)))
        model+=block(int(nROI/2), int(nROI/4))+block(int(nROI/4), nPattern)
        self.model = nn.Sequential(*model)

    def forward(self, input_y):
        zt = self.model(input_y)
        return zt

class LDecompose(nn.Module):
    def __init__(self, nPattern, nROI):
        super(LDecompose, self).__init__()
        model = []
        self.nPattern = nPattern
        self.nROI = nROI
        def block(in_layer, out_layer, normalize=True):
            layers=[]
            layers.append(nn.LeakyReLU(0.2, inplace=False))
            layers.append(nn.Linear(in_layer, out_layer))
            return layers
        model+=block(int(nROI), int(nPattern*nROI))
        self.model = nn.Sequential(*model)

    def forward(self, input_y):
        z_decompose = self.model(input_y)
        return [z_decompose[:,self.nROI*i:self.nROI*(i+1)] for i in range(self.nPattern)]


class LDiscriminator(nn.Module):
    def __init__(self, nROI, nPattern):
        super(LDiscriminator, self).__init__()
        self.model = nn.Sequential(
            nn.Linear(nROI, int(nROI/2),bias=True),
            nn.LeakyReLU(0.2, inplace=True),
            nn.Linear(int(nROI/2), int(nROI/4),bias=True),
            nn.LeakyReLU(0.2, inplace=True),
            nn.Linear(int(nROI/4), 2,bias=True),
        )
    def forward(self, input_y):
        pred = self.model(input_y)
        return pred





import os
import time
import numpy as np
from tqdm import tqdm
from torch.autograd import Variable
from .model import SurrealGAN
from .utils import Covariate_correction, Data_normalization, parse_train_data, parse_validation_data

__author__ = "Zhijian Yang"
__copyright__ = "Copyright 2019-2020 The CBICA & SBIA Lab"
__credits__ = ["Zhijian Yang"]
__license__ = "See LICENSE file"
__version__ = "0.1.0"
__maintainer__ = "Zhijian Yang"
__email__ = "zhijianyang@outlook.com"
__status__ = "Development"

class dotdict(dict):
    """dot.notation access to dictionary attributes"""
    __getattr__ = dict.get
    __setattr__ = dict.__setitem__
    __delattr__ = dict.__delitem__

class Surreal_GAN_train():
    
    def __init__(self, npattern, final_saving_epoch, max_epoch, recons_loss_threshold, mono_loss_threshold, \
        lam=0.2, zeta=80, kappa=80, gamma=6, mu=500, eta=6, batchsize=25, lipschitz_k = 0.5, \
        beta1 = 0.5, lr = 0.0002, max_gnorm = 100, eval_freq = 25, save_epoch_freq = 5, print_freq = 1000, saving_freq = 1000):
        self.opt=dotdict({})
        self.opt.npattern = npattern
        self.opt.final_saving_epoch = final_saving_epoch
        self.opt.max_epoch = max_epoch
        self.opt.recons_loss_threshold = recons_loss_threshold
        self.opt.mono_loss_threshold = mono_loss_threshold
        self.opt.lam = lam
        self.opt.mu = mu
        self.opt.zeta = zeta
        self.opt.kappa = kappa
        self.opt.gamma = gamma
        self.opt.eta = eta
        self.opt.batchsize = batchsize
        self.opt.lipschitz_k = lipschitz_k
        self.opt.beta1 = beta1
        self.opt.lr = lr
        self.opt.max_gnorm = max_gnorm
        self.opt.save_epoch_freq = save_epoch_freq
        self.opt.print_freq = print_freq
        self.opt.eval_freq = eval_freq
        self.opt.saving_freq = saving_freq

    def print_log(self, result_f, message):
        result_f.write(message+"\n")
        result_f.flush()
        print(message)

    def format_log(self, epoch, epoch_iter, measures, t, prefix=True):
        message = '(epoch: %d, iters: %d, time: %.3f) ' % (epoch, epoch_iter, t)
        if not prefix:
            message = ' ' * len(message)
        for key, value in measures.items():
            message += '%s: %.4f ' % (key, value)
        return message

    def parse_data(self, data, covariate, random_seed, data_fraction):
        cn_train_dataset, pt_train_dataset, correction_variables, normalization_variables = parse_train_data(data, covariate, random_seed, data_fraction, self.opt.batchsize)
        cn_eval_dataset, pt_eval_dataset = parse_validation_data(data, covariate, correction_variables, normalization_variables)
        self.opt.nROI = pt_eval_dataset.shape[1]
        self.opt.n_val_data = pt_eval_dataset.shape[0]
        return cn_train_dataset, pt_train_dataset, cn_eval_dataset, pt_eval_dataset, correction_variables, normalization_variables


    def train(self, data, covariate, save_dir, random_seed=0, data_fraction=1, verbose=True, independent_ROI = True):
        if verbose: 
            result_f = open("%s/results.txt" % save_dir, 'w')

        cn_train_dataset, pt_train_dataset, eval_X, eval_Y, correction_variables, normalization_variables = self.parse_data(data, covariate, random_seed, data_fraction)
        self.opt.correction_variables = correction_variables
        self.opt.normalization_variables = normalization_variables
        
        # create_model
        model = SurrealGAN()
        model.create(self.opt)

        total_steps = 0
        print_start_time = time.time()
        criterion_loss_list = [[0 for _ in range(2)] for _ in range(3)]                    ##### number of consecutive epochs with aq and cluster_loss < threshold
        predicted_label_past = np.zeros(self.opt.n_val_data)

        if self.opt.final_saving_epoch%self.opt.saving_freq == 0:
            save_epoch = [i * self.opt.saving_freq for i in range(2,self.opt.final_saving_epoch//self.opt.saving_freq+1)]
        else:
            save_epoch = [i * self.opt.saving_freq for i in range(2,self.opt.final_saving_epoch//self.opt.saving_freq+1)]+[self.opt.final_saving_epoch]
        save_epoch_index = 0 
        save_dir = os.path.join(save_dir,'repetition'+str(random_seed))
        if not os.path.exists(save_dir):
            os.makedirs(save_dir)
        
        if not verbose:
            pbar = tqdm(total = self.opt.max_epoch)
        for epoch in range(1, self.opt.max_epoch + 1):
            if not verbose:
                pbar.update(1)
            epoch_start_time = time.time()
            epoch_iter = 0
            for i, pt_data in enumerate(pt_train_dataset):
                cn_data = cn_train_dataset.next()
                real_X, real_Y = Variable(cn_data['x']), Variable(pt_data['y'])

                total_steps += self.opt.batchsize
                epoch_iter += self.opt.batchsize

                losses = model.train_instance(real_X, real_Y)

                if total_steps % self.opt.print_freq == 0:
                    t = (time.time() - print_start_time) / self.opt.batchsize
                    if verbose:
                        self.print_log(result_f, self.format_log(epoch, epoch_iter, losses, t))
                        print_start_time = time.time()

            if epoch % self.opt.save_epoch_freq == 0 and verbose:
                self.print_log(result_f, 'saving the model at the end of epoch %d.' % (epoch))
                model.save(save_dir,'latest')

            if epoch % self.opt.eval_freq == 0:
                t = time.time()

                loss_names = ['loss_recons','loss_mono']
                for _ in range(2):
                    criterion_loss_list[_].append(losses[loss_names[_]])

                for _ in range(2):
                    criterion_loss_list[_].pop(0) 

                t = time.time() - t
                res_str_list = ["[%d], TIME: %.4f" % (epoch, t)]

                if max(criterion_loss_list[0]) < self.opt.recons_loss_threshold \
                    and max(criterion_loss_list[1])<self.opt.mono_loss_threshold and epoch > save_epoch[save_epoch_index]:
                    model.save(save_dir, str(save_epoch[save_epoch_index])+'_epoch_model')
                    save_epoch_index+=1
                    res_str_list += ["*** Saving Criterion Satisfied ***"]
                    res_str = "\n".join(["-"*60] + res_str_list + ["-"*60])
                    if verbose:self.print_log(result_f, res_str)
                    if save_epoch_index==len(save_epoch):
                        if not verbose: pbar.close()
                        if verbose: result_f.close()
                        return True

                res_str = "\n".join(["-"*60] + res_str_list + ["-"*60])
                res_str_list += ["*** Max epoch reached and criterion not satisfied ***"]
                if verbose:  
                    self.print_log(result_f, res_str)
        if verbose: result_f.close()
        if not verbose: pbar.close()
        return False



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="SurrealGAN",
    version="0.0.5",
    author="zhijian.yang",
    author_email="zhijianyang@outlook.com",
    description="A python implementation of Surreal-GAN for semisupervised representation learning",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/zhijian-yang/SurrealGAN",
    packages=setuptools.find_packages(),
    classifiers=(
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ),
    install_requires=[         
        'numpy>=1.22.3',
        'tqdm>=4.50.2',
        'torch==1.10.2',
        'scikit-learn>=0.24.2',
        'scipy>=1.8.0',
        'pandas>=1.4.2',
        'lifelines>=0.26.3'
    ]
)

